package de.hs_mannheim.informatik.spreadsheet;

import javax.naming.ldap.spi.LdapDnsProviderResult;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.Scanner;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * A simplified spreadsheet class for the PR1 programming lab at Hochschule Mannheim.
 * One aspect worth mentioning is that it only supports long numbers, not doubles.
 *
 * @author  Oliver Hummel & Victor Waitz
 */
public class Spreadsheet {
	Cell[][] cells;

	/**
	 * Constructor that creates a Spreadsheet of size rows * cols.
	 * @param rows number of rows
	 * @param cols number of columns
	 */
	public Spreadsheet(int rows, int cols) {

		if (rows > 99){
			System.out.printf("Row size %d is bigger then 99, value will be set to 99! %n", rows);
			rows = 99;
		}
		if (cols > 26){
			System.out.printf("Col size %d is bigger then 26, value will be set to 26! %n", cols);
			cols = 26;
		}

		cells = new Cell[rows][cols];

		for (int r = 0; r < rows; r++)
			for (int c = 0; c < cols; c++)
				cells[r][c] = new Cell();
	}

	// -----
	// retrieve or change values of cells

	/**
	 * Private method to get a value from a cell in a spreadsheet.
	 * @param row The row of the cell.
	 * @param col The column of the cell.
	 * @return The value of the cell.
	 */
	private String get(int row, int col) {
		return cells[row][col].getValue();
	}

	/**
	 * Public method to get a value from a cell in a spreadsheet.
	 * @param cellName The name of the cell.
	 * @return The value of the cell.
	 */
	public String get(String cellName) {
		cellName = cellName.toUpperCase();
		return get(getRow(cellName), getCol(cellName));
	}

	/**
	 * Private methode to put a value into a cell in a spreadsheet.
	 * @param row The row of the cell.
	 * @param col The column of the cell.
	 * @param value The value to put into the cell.
	 * @return Nothing.
	 */
	private void put(int row, int col, String value) {
		if (!value.startsWith("="))
			cells[row][col].setValue(value);
		else {
			System.out.printf("Formula: %s %n", value);
			cells[row][col].setFormula(value);
			evaluateCell(row, col);
		}
	}

	/**
	 * Public method to put a value into a cell in a spreadsheet.
	 * Calls private methode put.
	 * @param cellName The name of the cell.
	 * @param value The value to put into the cell.
	 * @return Nothing.
	 */
	public void put(String cellName, String value) {
		cellName = cellName.toUpperCase();
		put(getRow(cellName), getCol(cellName), value);
	}

	/**
	 * Method to get the column of a cell in a spreadsheet.
	 * @param cellName The name of the cell.
	 * @return The column of the cell.
	 */
	public int getCol(String cellName) {
		return cellName.charAt(0) - 'A';
	}

	/**
	 * Method to get the row of a cell in a spreadsheet.
	 * @param cellName The name of the cell.
	 * @return The row of the cell.
	 */
	public int getRow(String cellName) {
		String row = "";
		for (int i = 1; i < cellName.length(); i++){
			row += cellName.charAt(i);
		}
		return Integer.parseInt(row) - 1;
	}

	// -----
	// business logic

	/**
	 * A method for reading in data from a CSV file.
	 * @param filePath The file to read.
	 * @param separator The char used to split up the input, e.g. a comma or a semicolon.
	 * @param startCellName The upper left cell where data from the CSV file should be inserted.
	 * @return Nothing.
	 * @exception IOException If path does not exist.
	 */
	public void readCsv(String filePath, String separator, String startCellName) throws FileNotFoundException {
		ArrayList<String> fileRows = new ArrayList<>();
		Scanner sc = new Scanner(new File(filePath));

		while (sc.hasNextLine()) {
			fileRows.add(sc.nextLine());
		}
		clearTable();

		ArrayList <int[]> formulas = new ArrayList<>();

		for (int rowI = 0; rowI < fileRows.size(); rowI++) {
			String row = fileRows.get(rowI);
			String[] cells = row.split(separator);

			for (int colI = 0; colI < cells.length; colI++) {
				String cellContent = cells[colI].toUpperCase();

				if (cellContent.startsWith("=")){
					formulas.add(new int[]{rowI, colI});
				}
				else {
					put(rowI, colI, cellContent);
				}

			}
		}

		for (int[] formulaPos : formulas){
			int rowI = formulaPos[0];
			int colI = formulaPos[1];
			System.out.printf("Formula at %d-%d %n", rowI, colI);

			String row = fileRows.get(rowI);
			String formulaToFill = row.split(separator)[colI];

			put(rowI, colI, formulaToFill);
		}

		sc.close();
	}

	/**
	 * A method for saving data to a CSV file.
	 * @param filePath The file to write.
	 * @return Nothing.
	 * @exception FileNotFoundException If path does not exist.
	 */
	public void saveCsv(String filePath) throws FileNotFoundException {
		PrintWriter out = new PrintWriter(filePath);

		for (Cell[] row : cells) {
			for (Cell cell : row) {
				if (!cell.getFormula().isEmpty())
					out.print("=" + cell.getFormula());
				else
					out.print(cell.getValue());

				if (cell != row[cells[0].length-1])
					out.print(",");
			}
			out.println();
		}

		out.close();
	}

	/**
	 * This method does the actual evaluation/calcluation of a specific cell
	 * @param col the col of the cell to be evaluated
	 * @param row the row of the cell to be evaluated
	 * @return Nothing.
	 */

	private void evaluateCell(int row, int col) {
		String formula = cells[row][col].getFormula();
		System.out.printf("Formel in eval: %s %n", formula);
		String result = "";

		if (formula.startsWith("SUMME("))  // e.g. SUMME(A3:A8)
			result = "" + sum(formula);

		else if (formula.startsWith("PRODUKT("))  // e.g. PRODUKT(A3:B9)
			result = "" + prod(formula);

		else if (formula.startsWith("MITTELWERT("))  // e.g. MITTELWERT(A3:A5)
			result = "" + avrg(formula);

		else if (formula.startsWith("STABW("))  // e.g. STABW(C6:D8) -> Standardabweichung
			result = "" + stdDevp(formula);

		else if (formula.startsWith("MIN("))  // e.g. MIN(C13:H13) -> größter Wert
			result = "" + min(formula);

		else if (formula.startsWith("MAX("))  // e.g. MAX(A1:A10) -> Standardabweichung
			result = "" + max(formula);

		else if (!formula.isEmpty()) {
			try {
				result = "" + calculate(formula);
			} catch(ArithmeticException ae) {
				result = "exc.";
			}
		}
		cells[row][col].setValue("" + result);
	}

	/**
	 * Method to extract the corner cells for a function formula.
	 * @param formula The formula to isolate the corners for
	 * @return The formulas corners.
	 */
	public String[] isolateFunctionCorners(String formula){
		//? isolate corners
		ArrayList<String> corners = new ArrayList<String>();

		Matcher m = Pattern.compile("[A-Z]\\d+").matcher(formula);
		while(m.find()) {
			String s = m.group();

			if (!s.isEmpty()){
				corners.add(s);
			}
		}
        return corners.toArray(new String[0]);
	}

	/**
	 * A methode to extract a block of cells used in a function formula,
	 * such as A1 and B3 for SUMME(A1:B3).
	 * Calls the isolateFunctionCorners methode to get the blocks corners.
	 * @param formula The formula to get the block for.
	 * @return The cell block.
	 */
	public String[] wholeFunctionBlock(String formula){
		String[] corners = isolateFunctionCorners(formula);

		String cornerOne = corners[0];
		char colOne = cornerOne.charAt(0);
		int rowOne = Integer.parseInt(cornerOne.substring(1));

		String cornerTwo = corners[1];
		char colTwo = cornerTwo.charAt(0);
		int rowTwo = Integer.parseInt(cornerTwo.substring(1));

		ArrayList<String> block = new ArrayList<>();

		for (int rowInt = rowOne; rowInt <= rowTwo; rowInt++) {
			for (char charCol = colOne; charCol <= colTwo; charCol++) {
				block.add(String.format("%c%d", charCol, rowInt));
			}
		}

		return block.toArray(new String[0]);
	}

	/**
	 * Method for calculating the sum of a rectangular block of cells, such as from A1 to B3.
	 * @param formula The formula to calculate the sum.
	 * @return The sum calculated.
	 */
	public long sum(String formula) {
		String[] block = wholeFunctionBlock(formula);
		long res = 0;

		for (String cellName : block){
			String cellValue = get(cellName);
			long value = cellValue.isEmpty() ? 0 : Long.parseLong(cellValue);
			res += value;
		}

		return res;
	}

	/**
	 * Method for calculating the product of a rectangular block of cells, such as from A1 to B3.
	 * @param formula The formula to calculate the product.
	 * @return The product calculated.
	 */
	public long prod(String formula){
		String[] block = wholeFunctionBlock(formula);

		long res = 1;
		boolean allEmpty = true;

		for (String cellName : block){
			String cellValue = get(cellName);
			if (!cellValue.isEmpty()){
				allEmpty = false;
			}
			long value = cellValue.isEmpty() ? 1 : Long.parseLong(cellValue);
			res *= value;
		}

		if (allEmpty){
			res = 0;
		}
		return res;
	}

	/**
	 * Method for calculating the average of a rectangular block of cells, such as from A1 to B3.
	 * @param formula The formula to calculate the average.
	 * @return The average calculated.
	 */
	public long avrg(String formula){
		String[] block = wholeFunctionBlock(formula);

		boolean allEmpty = true;
		long res = 0;
		int counter = 0;

		for (String cellName : block){
			String cellValue = get(cellName);
			if (cellValue.isEmpty()){
				continue;
			}
			allEmpty = false;

			res += Long.parseLong(cellValue);
			counter++;
		}
		if (allEmpty){
			System.out.println("Division by zero! Value set to 0");
			return 0;
		}

		res = res/counter;
		return res;
	}

	/**
	 * Method for calculating the standard deviation of a rectangular block of cells, such as from A1 to B3.
	 * Calls the average method.
	 * @param formula The formula to calculate the standard deviation.
	 * @return The standard deviation calculated.
	 */
	public long stdDevp(String formula){
		String[] block = wholeFunctionBlock(formula);

		String[] notEmptyValues = extractNotEmptyCells(block);

		if(notEmptyValues.length < 2){
			System.out.println("Division by zero! Value set to 0");
			return 0;
		}

		long meanValue = avrg(formula);
		long squaredDiffsSum = 0;

		for (String value : notEmptyValues){
			long difference = Long.parseLong(value) - meanValue;
			squaredDiffsSum += (long) Math.pow(difference, 2);
		}

		long variance = squaredDiffsSum / (notEmptyValues.length - 1);
		return (long) Math.sqrt(variance);
	}

	/**
	 * Method for selecting the minimum of a rectangular block of cells, such as from A1 to B3.
	 * @param formula The formula to select the minimum from.
	 * @return The minimum selected.
	 */
	public long min(String formula){
		String[] block = wholeFunctionBlock(formula);

		String firstElement = "";
		long currentSmallest = 0;
		for (int i = 0; i < block.length; i++) {
			String cellValue = get(block[i]);
			if (cellValue.isEmpty()) {
				continue;
			}
			if (firstElement.isEmpty()){
				firstElement = get(block[i]);
				currentSmallest = Long.parseLong(firstElement);
			}

			long value = Long.parseLong(cellValue);
			if (value < currentSmallest){
				currentSmallest = value;
			}
		}
		return currentSmallest;
	}

	/**
	 * Method for selecting the maximum of a rectangular block of cells, such as from A1 to B3.
	 * @param formula The formula to select the maximum from.
	 * @return The maximum selected.
	 */
	public long max(String formula){
		String[] block = wholeFunctionBlock(formula);

		String firstElement = "";
		long currentBiggest = 0;
		for (int i = 0; i < block.length; i++) {
			String cellValue = get(block[i]);
			if (cellValue.isEmpty()) {
				continue;
			}

			if (firstElement.isEmpty()){
				firstElement = get(block[i]);
				currentBiggest = Long.parseLong(firstElement);
			}

			long value = Long.parseLong(cellValue);
			if (value > currentBiggest){
				currentBiggest = value;
			}
		}
		return currentBiggest;
	}

	/**
	 * Method to extract only cells which are not empty from a previously calculated block from a
	 * function formula.
	 * @param block Block with empty and not empty cells.
	 * @return Block with only not empty cells.
	 */
	public String[] extractNotEmptyCells(String[] block){
		ArrayList<String> notEmptyValues = new ArrayList<String>();

		for (String cell : block) {
			String cellValue = get(cell);

			if (!cellValue.isEmpty()){
				notEmptyValues.add(cellValue);
			}
		}
		return notEmptyValues.toArray(new String[0]);
	}

	/**
	 *
	 * This method calculates the result of a "normal" algebraic expression. It only needs to support
	 * expressions like =B4 or =2+A3-B2, i.e. only with int numbers and other cells and with plus,
	 * minus, times, split only. An expression always starts with either a number or a cell name. If it 
	 * continues, it is guaranteed that this is followed by an operator and either a number or a 
	 * cell name again. It is NOT required to implement dot before dash or parentheses in formulas.
	 * @param formula The expression to be evaluated.
	 * @return The result calculated.
	 */
	public long calculate(String formula) throws ArithmeticException {
		long res = 0;

		ArrayList<String> formulaElements = extractExpressionElements(formula);

		String firstElement = formulaElements.get(0);

		if (isValueCellName(firstElement).equals("cellName")){
			String cellValue = get(firstElement);
			res = cellValue.isEmpty() ? 0 : Long.parseLong(cellValue);
		}
		else {
			res = Long.parseLong(firstElement);
		}

		for (int i = 1; i < formulaElements.size(); i = i+2) {
			res = singleCalculation(res, formulaElements.get(i+1), formulaElements.get(i));
		}

		return res;
	}

	/**
	 * Method to extract every single element of an expression and stores them as a separate element in the
	 * String ArrayList formulaElements and returns it.
	 * An expression element is either a number, a cell name or an operator.
	 * @param formula The formula expression to extract elements from.
	 * @return The elements as single elements.
	 */
	public static ArrayList<String> extractExpressionElements(String formula){
		Matcher m = Pattern.compile("([A-Z][0-9]*)|[-\\+\\*/]|[0-9]*").matcher(formula);

		ArrayList<String> formulaElements = new ArrayList<String>();

		while (m.find()) {
			String s = m.group();
			if (!s.isEmpty()) {
				formulaElements.add(s);
			}
		}

		return formulaElements;
	}

	/**
	 * Method to calculate a single calculation. A single calculation always uses one operator and one value part.
	 * The value part can be a cell name or a value. The calculation takes the given previous solution and uses the given
	 * operator and the given value part to calculate and return a new solution.
	 * @param res The previous result of a calculation.
	 * @param element The element to calculate to the previous result.
	 * @param operator The operator used to calculate previous result and element.
	 * @return The new result.
	 */
	private long singleCalculation(long res, String element, String operator){
		String elementType = isValueCellName(element);

		if (elementType.equals("cellName")){
			element = get(element);
		}
		long value = element.isEmpty() ? 0 : Long.parseLong(element);


		switch (operator) {
			case "+":
				res += value;
				break;
			case "-":
				res -= value;
				break;
			case "*":
				res *= value;
				break;
			case "/":
				if (value == 0){
					throw new ArithmeticException("Division by zero");
				}
				res /= value;
				break;
		}
		return res;
	}

	/**
	 * Methode to determine if a value part is either a cell name or a value.
	 * Has the possibility to determine floating point numbers for the future.
	 * @param sToCheck The string to decide.
	 * @return The String containing the information if it is a cell name or a value.
	 */
	public String isValueCellName(String sToCheck){
		//? 7 -> "value", A1 -> "cellName", 7A -> "invalid"

		//? For floating point numbers
		// String digitCheckRegex = "-?\\d+(\\.\\d+)?";
		//? For integers
		String digitCheckRegex = "-?\\d+";
		if (sToCheck.matches(digitCheckRegex)){
			return "value";
		}

		//? For cell names
		String cellNameCheckRegex = "[A-Z][0-9]*";
		if (sToCheck.matches(cellNameCheckRegex)){
			return "cellName";
		}

		return "invalid";
	}

	// -----

	/**
	 * Returns the number of rows in the spreadsheet.
	 * @return amount rows.
	 */
	public int getRowsLCount() {
		return cells.length;
	}

	/**
	 * Returns the number of columns in the spreadsheet.
	 * @return amount columns.
	 */
	public int getColsCount() {
		return cells[0].length;
	}

	/**
	 * Clears the spreadsheet, it fills every value and formula to zero.
	 * @return Nothing.
	 */
	public void clearTable() {
		int rows = getRowsLCount();
		int cols = getColsCount();

		for (int r = 0; r < rows; r++) {
			for (int c = 0; c < cols; c++) {
				cells[r][c].setValue("");
				cells[r][c].setFormula("");
			}
		}
	}

	/**
	 * Loads a special table for testing purposes.
	 * @return Nothing.
	 */
	public void loadSpecialTable(){
		clearTable();

		put("A1", "1");
		put("A2", "2");
		put("A3", "3");
		put("A4", "4");
		put("B1", "5");
		put("B2", "6");
		put("B3", "7");
		put("B4", "8");
		put("K1", "200");
		put("K11", "100");
		put("J4", "4");
		put("I10", "69");
		put("F5", "3");
		put("G4", "2");
		put("G5", "3");
	}

	/**
	 * Returns a string representation of the spreadsheet.
	 * @return The spreadsheet as a string.
	 */
	public String toString() {
		StringBuilder sb = new StringBuilder();

		sb.append("    ");
		for (int i = 0; i < cells[0].length; i++) {
			sb.append("  " + (char)('A'+ i) + "  | ");
		}

		int rc = 1;
		for (Cell[] r : cells) {
			sb.append(System.lineSeparator());
			sb.append(String.format("%2s", rc++) + ": ");

			for (Cell c : r) {
				sb.append(c + " | ");
			}

		}
		return sb.toString();
	}
}